/*  $Id: solo_mm.cc,v 1.11 1999/07/16 20:18:48 shaggy Exp $
 *      SOLO memory management routines
 *      relative path: ./lib/boot/m-pc
 */

#include <libc_port.h>
#include <stdx1.h>
#include <defines>
#include <boot/bootparam.h>
#include <boot/fileinfo.h>


#include "libsolo.h"


// Note:
//	these memory allocation routines probably don't need to
//	be very efficient, since they are only supposed to be used
//	by utilities that run once, then exit, so even if they waste
//	memory, it's only very temporary.
//	Any real program (such as an operating system kernel)
//	should have its own memory allocation routines.
//	These are just here for convenience...
//	so, once memory fragments with these routines, it might never rejoin...
//	Use sparingly... 

//	XXX - some consistency checks could be made on these memory
//	blocks to detect corruption--- such as checking that 
//	'next' pointers are always greater, and 'prev' pointers are
//	always less, as well as some checks on the sizes

//	XXX - all of this code depends upon the free list staying in
//	order, as well as the main 'freelist' structure being less
//	in address than any of the nodes allocated


struct MemBlock {
	vaddr_t	size;
	union {
		MemBlock *next;
		MemBlock *prev;
	} ptr;
};


static MemBlock freelist;


void*
solo_calloc(size_t num, size_t size)
{
	void *ptr = solo_malloc(num*size);
	memset(ptr, 0, num*size);
	return ptr;
}


void*
solo_malloc(size_t size)
{
	MemBlock *fp_prev = &freelist;
	MemBlock *fp;

	// assert(is_pow2(sizeof(MemBlock)));
	size = align(size, sizeof(MemBlock));

	while((fp = fp_prev->ptr.next) && size > fp->size)
		fp_prev = fp_prev->ptr.next;

	if(!fp) 
		return 0;

	void *adr = (void*)((vaddr_t)fp + sizeof(MemBlock));
	size_t freesize = fp->size;

	if(fp->size - size > sizeof(MemBlock)) {
		MemBlock *newfp = (MemBlock*)((vaddr_t)adr + size);
		newfp->ptr.next = fp->ptr.next;
		newfp->size = freesize - (size+sizeof(MemBlock));
		fp_prev->ptr.next = newfp;
		fp->size = size;
	} else 
		fp_prev->ptr.next = fp->ptr.next;
	fp->ptr.prev = fp_prev;
	return adr;
}


void*
solo_realloc(void *ptr, size_t size)
{
	MemBlock *fp = (MemBlock*)((vaddr_t)ptr - sizeof(MemBlock));
	if(fp->size >= size)
		return ptr;
	void *newp = solo_malloc(size);
	memcpy(newp, ptr, size);
	solo_free(ptr);
	return newp;
}


void
solo_free(void *ptr)
{
	if(!ptr) 
		return;
	MemBlock *fp = (MemBlock*)((vaddr_t)ptr - sizeof(MemBlock));
	MemBlock *fp_prev = fp->ptr.prev;
	while(fp_prev->ptr.prev && fp_prev->ptr.prev < fp_prev)
		fp_prev = fp_prev->ptr.prev;
	while(fp_prev->ptr.next && fp_prev->ptr.next < fp)
		fp_prev = fp_prev->ptr.next;
	fp->ptr.next = fp_prev->ptr.next;
	fp_prev->ptr.next = fp;
	// XXX - should do block rejoining here, but we have
	// to be careful to fix any 'prev' pointers that may point
	// at the block header we're eliminating
}



extern "C" bool _bootparam_init();

extern "C" bool
_solo_mm_init()
{
	static bool initialized = false;
	static bool initializing = false;
	BootParamPtr bootparam;

	if(initialized)
		return true;
	if(initializing)
		return false;
	initializing = true;

	_bootparam_init();

	freelist.ptr.next = (MemBlock*)bootparam->lomem;
	freelist.ptr.next->size =
		bootparam->himem-bootparam->lomem - sizeof(MemBlock);
	freelist.ptr.next->ptr.next = 0;
	
	initialized = true;
	initializing = false;
	return true;
}

// Global constructor:
bool _solo_mm_initialized = _solo_mm_init();

