/*  $Id: Vec.hh,v 1.8 2000/01/11 07:32:38 shaggy Exp $
 *      relative path: ./lib/c++
 */

#ifndef __Vec_hh__
#define __Vec_hh__

#ifndef _POSIX_SOURCE
# define _POSIX_SOURCE
#endif

#include <stdx1.h>
#include <sys/types.h>

// ADR_LEN(d) - generates a pointer and size comma-separated pair 
// STR_LEN(d) - generates a pointer and size pair for a string
// CSTR_LEN(d) - generates a pointer and size pair for a constant string
// STR(s) - generates Str object from a character string
// CSTR(s) - generates CStr object from a constant character string
// _STR(s) - generates Str structure from a character string
// _CSTR(s) - generates CStr structure from a constant character string
// VEC(t,v) - generates Vec<t> object from a vector
// CVEC(t,v) - generates CVec<t> object from a fixed vector
// _VEC(t, v) - generates Vec<t> structure from a vector
// _CVEC(t, v) - generates CVec<t> structure from a constant vector
// INLINE_VEC(t) - creates inline instantiations of vector code
//		   for the specified type 
//
// Note: using the structure forms (i.e. _CSTR(s)) instead of
//	 the class forms whenever possible in static arrays
//	 will produce smaller code, since the structures will
//	 all be automatically generated in the data section,
//	 and won't require constructor calls.
//	One problem with C++ classes in general (if they
//	have constructor calls) is that even if the call
//	is constant and could be optimized, (in GNU g++ at least)
//	a constructor call is made for every initialization,
//	which is a lot of overhead in the case of large arrays.
	

// XXX - originally, we stored the length (size of the array)
//	 which meant that when converting between types,
//	 that size had to be modified, and looking at
//	 an unknown type, the memory size wasn't known.
//	 Now, we use the byte count as the stored number,
//	 and compute the array size when needed.
//	(hopefully, this isn't a mistake)

#define ADR_LEN(d)	&d, sizeof(d)
#define ARR_LEN(a)	(a), sizeof(a)
#define STR_LEN(s)	(s), strlen(s)
#define CSTR_LEN(s)	(s), sizeof(s)-1
#define _STR(s)		(_Str){DSTR_LEN(char, s)}
#define _CSTR(s)	(_CStr){CSTR_LEN(const char, s)}
#define _VEC(t,v)	(_Vec<t>){v, sizeof(v)}
#define _CVEC(t,v)	(_CVec<t>){v, sizeof(v)}
#define STR(s)		Str(STR_LEN(s))
#define CSTR(s)		CStr(CSTR_LEN(s))
#define VEC(t,v)	Vec<t>(v, sizeof(v))
#define CVEC(t,v)	CVec<t>(v, sizeof(v))
#define INLINE_VEC(T) \
		inline template class _CVec<T>; \
		inline template class _Vec<T>; \
		inline template class CVec<T>; \
		inline template class Vec<T>;

// Note that casts can be done on ADR_LEN, STR_LEN, and CSTR_LEN, like:
//	uchar_write( (uchar)CSTR_LEN("testing") )
// (since the cast applies to the first item)


template <class T> struct _CVec {
	const T *_ptr;
	size_t _len;
	size_t count() const { return _len/sizeof(T); }
	size_t size() const { return _len; }
	const T* ptr() const { return _ptr; }
	operator const T* () const { return _ptr; }
	const T& operator[](size_t n) const { return _ptr[n]; }
	_CVec& operator++() { _ptr++; _len -= sizeof(T); return *this; }
	void operator++(int) { _ptr++; _len -= sizeof(T); }
	_CVec& operator+=(ssize_t n) { _ptr += n; _len -= n*sizeof(T); return *this; }
};


template <class T> struct _Vec {
	T *_ptr;    
	size_t _len; 
	size_t count() const { return _len/sizeof(T); }
	size_t size() const { return _len; } 
	T* ptr() const { return _ptr; }
	operator T* () const { return _ptr; }
	operator const T* () const { return _ptr; }
	T& operator[](size_t n) const { return _ptr[n]; }
	operator _CVec<T>& () { return (_CVec<T>)(*this); }
	_Vec& operator++() { _ptr++; _len -= sizeof(T); return *this; }
	void operator++(int) { _ptr++; _len -= sizeof(T); }
	_Vec& operator+=(ssize_t n) { _ptr += n; _len -= n*sizeof(T); return *this; }
}; 


// XXX - should just be able to have default types of 'void'
//	(i.e.  CVec<>  where 'class T = void' )
// maybe once gcc catches up to the standard??


struct _CVec_void {
	const void *_ptr;
	size_t _len;
	size_t size() const { return _len; }
	const void* ptr() const { return _ptr; }
	operator const void* () const { return _ptr; }
};

struct _Vec_void {
	void *_ptr;    
	size_t _len; 
	size_t size() const { return _len; } 
	void* ptr() { return _ptr; }
	operator void* () { return _ptr; }
	operator const void* () const { return _ptr; }
	operator _CVec_void& () const { return reinterpret_cast<_CVec_void&>(
			const_cast<_Vec_void&>(*this)); }
}; 

typedef _CVec_void _CVecAny;
typedef _Vec_void _VecAny;


// Note that direct instantiation of these is not allowed

class CVecAny : public _CVecAny {
private:
	CVecAny() {}
};

class VecAny : public _VecAny {
private:
	VecAny() {}
public:
	operator CVecAny&() const { return *(CVecAny*)this; }
};

// XXX - these should not be necessary if Vec<void> could be 
// 	declared (and it will be, in the future)

class CVec_void;

class Vec_void : public _Vec_void {
public:
	Vec_void() { _ptr = 0; _len = 0; }
	Vec_void(const _VecAny &dv) { _ptr = dv._ptr; _len = dv._len; }
	Vec_void(const Vec_void &dv) { _ptr = dv._ptr; _len = dv._len; }
	Vec_void(void *ptr, size_t len) { _ptr = ptr; _len = len; }
	operator CVec_void& () { return *(CVec_void*)this; }
	explicit Vec_void(const VecAny &va) { _ptr = va._ptr; _len = va._len; }
	operator VecAny&() { return *(VecAny*)this; }
	operator CVecAny&() { return *(CVecAny*)this; }
};

class CVec_void : public _CVec_void {
public:
	CVec_void() { _ptr = 0; _len = 0; }
	CVec_void(const _CVec_void &cv) { _ptr = cv._ptr; _len = cv._len; }
	CVec_void(const _Vec_void &dv) { _ptr = dv._ptr; _len = dv._len; }
	CVec_void(const CVec_void &cv) { _ptr = cv._ptr; _len = cv._len; }
	CVec_void(const Vec_void &dv) { _ptr = dv._ptr; _len = dv._len; }
	CVec_void(const void *ptr, size_t len) { _ptr = ptr; _len = len; }
	explicit CVec_void(const VecAny &va)
			{ _ptr = va._ptr; _len = va._len; }
	explicit CVec_void(const CVecAny &va)
			{ _ptr = va._ptr; _len = va._len; }
	operator CVecAny&() { return *(CVecAny*)this; }
};


template <class T> class Vec : public _Vec<T> {
public:
	Vec<T>() { _ptr = 0; _len = 0; }
	Vec<T>(const _Vec<T> &dv) { _ptr = dv._ptr; _len = dv._len; }
	Vec<T>(const Vec<T> &dv) { _ptr = dv._ptr; _len = dv._len; }
	Vec<T>(T *ptr, size_t len) { _ptr = ptr; _len = len; }
	Vec<T>(const Vec<T> &dv, size_t pos)
		{ _ptr = dv._ptr+pos; _len = dv._len - pos*sizeof(T); }
	Vec<T>(const Vec<T> &dv, size_t pos, size_t count)
		{ _ptr = dv._ptr+pos; _len = count*sizeof(T); }
	operator _CVec<T>& () { return *(_CVec<T>*)this; }
	explicit Vec<T>(const VecAny &va) { _ptr = (T*)va._ptr; _len = va._len; }
	Vec<T>(T &item) { _ptr = &item; _len = 1; }
	operator VecAny&() { return *(VecAny*)this; }
	operator CVecAny&() { return *(CVecAny*)this; }
};

template <class T> class CVec : public _CVec<T> {
public:
	CVec<T>() { _ptr = 0; _len = 0; }
	CVec<T>(const _CVec<T> &cv) { _ptr = cv._ptr; _len = cv._len; }
	CVec<T>(const _Vec<T> &dv) { _ptr = dv._ptr; _len = dv._len; }
	CVec<T>(const CVec<T> &cv) { _ptr = cv._ptr; _len = cv._len; }
	CVec<T>(const Vec<T> &dv) { _ptr = dv._ptr; _len = dv._len; }
	CVec<T>(const T *ptr, size_t len) { _ptr = ptr; _len = len; }
	CVec<T>(const CVec<T> &dv, size_t pos)
		{ _ptr = dv._ptr+pos; _len = dv._len - pos*sizeof(T); }
	CVec<T>(const CVec<T> &dv, size_t pos, size_t count)
		{ _ptr = dv._ptr+pos; _len = count*sizeof(T); }
	CVec<T>(const Vec<T> &dv, size_t pos)
		{ _ptr = dv._ptr+pos; _len = dv._len - pos*sizeof(T); }
	CVec<T>(const Vec<T> &dv, size_t pos, size_t count)
		{ _ptr = dv._ptr+pos; _len = count*sizeof(T); }
	explicit CVec<T>(const VecAny &va)
			{ _ptr = (const T*)va._ptr; _len = va._len; }
	explicit CVec<T>(const CVecAny &va)
			{ _ptr = (const T*)va._ptr; _len = va._len; }
	CVec<T>(const T &item)
			{ _ptr = &item; _len = 1; }
	operator CVecAny&() { return *(CVecAny*)this; }
};

// Note that 'signed char' and 'unsigned char'
// are *not* included as builtin char types, because
// when the signed-ness of a char type is being depended upon,
// it is generally being used as an integer, not a character.

#ifndef FOR_ALL_BUILTIN_CHAR_TYPES
#define FOR_ALL_BUILTIN_CHAR_TYPES(x)   \
                x(char) x(wchar_t)
#endif

#ifndef FOR_ALL_BUILTIN_INT_TYPES
#define FOR_ALL_BUILTIN_INT_TYPES(x)    \
                x(int8) x(uint8) \
                x(int16) x(uint16) \
                x(int32) x(uint32) \
                x(int64) x(uint64)
#endif

#ifndef FOR_ALL_BUILTIN_FLOAT_TYPES
#define FOR_ALL_BUILTIN_FLOAT_TYPES(x)  \
                x(float) x(double) x(long double)
#endif

#ifndef FOR_ALL_BUILTIN_TYPES
#define FOR_ALL_BUILTIN_TYPES(x)        FOR_ALL_BUILTIN_CHAR_TYPES(x)   \
                                        FOR_ALL_BUILTIN_INT_TYPES(x)    \
                                        FOR_ALL_BUILTIN_FLOAT_TYPES(x)
#endif


FOR_ALL_BUILTIN_TYPES( INLINE_VEC )
// XXX - is this necessary???
// Just to be sure everything is inlined: 
// FOR_ALL_BUILTIN_TYPES( INLINE_VEC )

typedef Vec<char> Str;
typedef CVec<char> CStr;
typedef Vec<wchar_t> WStr;
typedef CVec<wchar_t> CWStr;


/////////////////////////////////////////////////////////////////////////////
// XXX - okay, the rest of this file is old stuff, being kept
// 	around only for reference until we're sure we don't need it any more



#ifdef old_stuff

#ifndef __NO_TEMPLATES__



#define STRUCT_VEC_named(T, CN, DN) \
			inline template class _CVec<T>; \
			inline template class _Vec<T>; \
			inline template class CVec<T>; \
			inline template class Vec<T>; \
			typedef _CVec<T> _##CN; \
			typedef _Vec<T> _##DN; \
			typedef CVec<T> CN; \
			typedef Vec<T> DN; 
#else

//  This code was needed before template support worked
//  correctly with optimized inline code...
//  (and it's now grossly out of date, so it needs major
//   changes if it's ever to be used again)


#define STRUCT_VEC_named(T, CN, DN)	\
	struct CN { \
		const T *ptr;    \
		size_t len;   \
		size_t size() const { return sizeof(T)*len; } \
	}; \
	struct DN { \
		T *ptr;    \
		size_t len;   \
		size_t size() const { return sizeof(T)*len; } \
		operator CN& () { return (CN)(*this); } \
	}; 

#ifdef unused
		operator CVec() \
			{ return (CVec){(const uchar*)ptr, sizeof(T)*len}; } \
		CN& operator=(CVec &cv) \
			{ ptr = (const T*)cv.ptr; \
				len = cv.len/sizeof(T); return *this; } \

		operator Vec() \
			{ return (Vec){(uchar*)ptr, sizeof(T)*len}; } \
		DN& operator=(Vec &dv) \
			{ ptr = (T*)dv.ptr; \
				len = dv.len/sizeof(T); return *this; } 
#endif


#define STRUCT_VEC(T)	STRUCT_VEC_named(T, CVec_##T, Vec_##T) 

#endif


#define STRUCT_VEC(T)	STRUCT_VEC_named(T, CVec_##T, Vec_##T) 


// Some basic types:

STRUCT_VEC(char)
STRUCT_VEC(uchar)
STRUCT_VEC(int)
STRUCT_VEC(uint)
STRUCT_VEC_named(char, CStr, DStr)

#endif

#endif
